@@ -2,13 +2,42 @@
 
 library(expm)
 
-source(system.file("test-tools.R", package = "expm"))## -> assertError(), rMat()
-source(system.file("demo", "exact-fn.R", package = "expm"))
+options(digits = 4, width = 90, keep.source = FALSE)
+
+mSource <- function(file, ...)
+    source(system.file(file, ..., package = "expm", mustWork=TRUE))
+mSource("test-tools.R")## -> assertError(), rMat(), .. doExtras
+mSource("demo", "exact-fn.R")
+doExtras
+
+re.nilA3 <- function(xyz, EXPMlist)
+{
+    stopifnot(is.list(EXPMlist))
+    r <- do.call(nilA3, as.list(xyz))
+    sapply(EXPMlist, function(Efn) relErr(r$expA, Efn(r$A)))
+}
+
+re.facMat <- function(n, EXPMlist, rFUN = rnorm, ...)
+{
+    stopifnot(is.list(EXPMlist))
+    r <- facMat(n, rFUN, ...)
+    vapply(EXPMlist, function(EXPM) {
+	ct <- system.time(E <- EXPM(r$A), gc = FALSE)[[1]]
+	c(relErr = relErr(r$expA, E), c.time = ct)
+    }, double(2))
+}
+
+re.m2ex3 <- function(eps, EXPMlist)
+{
+    stopifnot(is.list(EXPMlist))
+    r <- m2ex3(eps)
+    sapply(EXPMlist, function(EXPM) relErr(r$expA, EXPM(r$A)))
+}
 
 set.seed(321)
 re <- replicate(1000,
-                c(re.nilA3(rlnorm(3),function(x)expm(x,"Pade")),
-                  re.nilA3(rnorm(3), function(x)expm(x,"Pade"))))
+                c(re.nilA3(rlnorm(3),list(function(x)expm(x,"Pade"))),
+                  re.nilA3(rnorm(3), list(function(x)expm(x,"Pade")))))
 
 summary(t(re))
 stopifnot(rowMeans(re) < 1e-15,
@@ -16,14 +45,13 @@ stopifnot(rowMeans(re) < 1e-15,
           apply(re, 1, quantile, 0.90) < 2e-15,
           apply(re, 1, max) < c(4e-14, 6e-15))
 
-cat('Time elapsed: ', (p1 <- proc.time()),'\n') # for ``statistical reasons''
+showProc.time()
 
-
-## Check *many* random nilpotent matrices:
+## Check *many* random nilpotent  3 x 3  matrices:
 set.seed(321)
 RE <- replicate(1000,
-                c(re.nilA3(rlnorm(3), function(x) expm(x, "Ward77")),
-                  re.nilA3(rnorm(3),  function(x) expm(x, "Ward77"))))
+                c(re.nilA3(rlnorm(3), list(function(x) expm(x, "Ward77"))),
+                  re.nilA3(rnorm(3),  list(function(x) expm(x, "Ward77")))))
 stopifnot(rowMeans(RE) < 1e-15,
           apply(RE, 1, quantile, 0.80) < 1e-16,
           apply(RE, 1, quantile, 0.90) < 2e-15,
@@ -35,8 +63,15 @@ cat("relErr(expm(.,Pade)) - relErr(expm(.,Ward77))  in Machine_eps units:\n")
 print(summary(c(re - RE)) / epsC)
 ##       Min.    1st Qu.     Median       Mean    3rd Qu.       Max.
 ## -0.6183442  0.0000000  0.0000000  1.3650410  0.1399719 94.9809161
+## nb-mm3; ditto lynne (both x64), 2014-09-11:
+##   Min. 1st Qu.  Median    Mean 3rd Qu.    Max.
+## -0.8422  0.0000  0.0000  0.0725  0.1067  1.2205
+## 32-bit [i686, florence, Linx 3.14.8-100.fc19..]:
+##  Min. 1st Qu.  Median    Mean 3rd Qu.    Max.
+## -0.62    0.00    0.00    1.36    0.14   95.93
 
-cat('Time elapsed: ',(p2 <- proc.time())-p1,'\n') # for ``statistical reasons''
+
+showProc.time()
 
 ###--- A second group --- where we know the diagonalization of A ---
 
@@ -48,36 +83,91 @@ if(!require("Matrix"))
 
 ## rMat() relies on Matrix::rcond():
 ## Now with the change default rcondMin, this "works"
-system.time(R40 <- rMat(40))
-system.time(R80 <- rMat(80))
+R40 <- rMat(40)
+R80 <- rMat(80)
+showProc.time()
 
 expm.safe.Eigen <- function(x, silent = FALSE) {
     r <- try(expm::expm(x, "R_Eigen"), silent = silent)
     if(inherits(r, "try-error")) NA else r
 }
 
+## the S4 generic
+Matrix::expm
+## the dgeMatrix method:
+expm.Matr.dge <- function(x) getDataPart(getMethod("expm", "dgeMatrix"))(Matrix::..2dge(x))
 expmList <-
-    list(Ward  = function(x) expm::expm(x, "Ward77"),
+    list(Matr = Matrix::expm,
+         Matr.d = expm.Matr.dge,
+         Ward  = function(x) expm::expm(x, "Ward77"),
 	 s.P.s = function(x) expm::expm(x, "Pade"),
 	 s.P.sO= function(x) expm::expm(x, "PadeO"),
-	 sPs.H08.= function(x) expm::expm.Higham08(x, balancing=FALSE),
-	 sPs.H08b= function(x) expm::expm.Higham08(x, balancing= TRUE),
+	 s.P.sRBS= function(x) expm::expm(x, "PadeRBS"),
+	 sPs.H08.= function(x) expm:: expm.Higham08(x, balancing=FALSE),
+	 sPs.H08b= function(x) expm:: expm.Higham08(x, balancing= TRUE),
+	 AmHi09.06= function(x) expm:::expm.AlMoHi09(x, p =  6),
+	 AmHi09.08= function(x) expm:::expm.AlMoHi09(x, p =  8),
+	 AmHi09.10= function(x) expm:::expm.AlMoHi09(x, p = 10),
+	 AmHi09.12= function(x) expm:::expm.AlMoHi09(x, p = 12),
+	 AmHi09.13= function(x) expm:::expm.AlMoHi09(x, p = 13),
 	 s.T.s = function(x) expm::expm(x, "Taylor"),
 	 s.T.sO= function(x) expm::expm(x, "TaylorO"),
 	 Eigen = expm.safe.Eigen,
 	 hybrid= function(x) expm::expm(x, "hybrid")
 	 )
-expmL.wo.E <- expmList[names(expmList) != "R_Eigen"]
 
 
 set.seed(12)
-re.facMat(20, expmList)
-fRE <- replicate(100, re.facMat(20, expmList))
+fRE <- replicate(if(doExtras) 100 else 20,
+                 re.facMat(20, expmList))
+cat("Number of correct decimal digits for facMat(20, rnorm):\n")
+summary(-log10(t(fRE["relErr",,])))
+
 
 ## Now look at that:
-boxplot(t(fRE), log="y", notch=TRUE,
+boxplot(t(fRE["relErr",,]), log="y", notch=TRUE, ylim = c(8e-16, 1e-8),
         main = "relative errors for 'random' eigen-ok 20 x 20 matrix")
 
+showProc.time()
+
+if(doExtras) {
+str(rf100 <- replicate(20, re.facMat(100, expmList)))
+print(1000*t(apply(rf100["c.time",,], 1, summary)))
+## lynne {Linux 2.6.34.7-56.fc13.x86_64 --- AMD Phenom II X4 925}:
+##          Min. 1st Qu. Median  Mean 3rd Qu. Max.
+## Ward       23      24   24.5  24.4    25.0   25
+## s.P.s     107     109  109.0 109.0   109.0  112
+## s.P.sO    188     190  191.0 192.0   193.0  198
+## s.P.sRBS   17      18   19.0  18.9    19.2   21
+## sPs.H08.   15      17   18.0  17.6    18.0   19
+## sPs.H08b   18      18   19.0  23.4    20.0  107
+## s.T.s      44      45   45.0  45.6    46.0   48
+## s.T.sO     96      98   99.0 100.0   100.0  116
+## Eigen      18      19   20.0  24.4    21.0  109
+## hybrid     40      42   42.0  47.1    44.0  133
+
+##--> take out the real slow ones for the subsequent tests:
+`%w/o%` <- function(x, y) x[!x %in% y] #--  x without y
+print(nms.swift <- names(expmList) %w/o%
+ c("s.P.s", "s.P.sO", "s.T.s", "s.T.sO"))
+expmL.swift <- expmList[nms.swift]
+
+set.seed(18) ## 12 replicates is too small .. but then it's too slow otherwise:
+rf400 <- replicate(12, re.facMat(400, expmL.swift))
+print(1000*t(apply(rf400["c.time",,], 1, summary)))
+## lynne:
+##          Min. 1st Qu. Median Mean 3rd Qu. Max.
+## Ward     1740    1790   1830 1820    1860 1900
+## s.P.sRBS 1350    1420   1440 1430    1450 1460
+## sPs.H08. 1020    1030   1130 1140    1210 1290
+## sPs.H08b 1120    1130   1220 1220    1300 1390
+## Eigen     962     977    989  992    1000 1030
+## hybrid   2740    2800   2840 2840    2890 2910
+
+showProc.time()
+
+}## if(doExtras)  only
+
 ## Now  try an example with badly conditioned "random" M matrix...
 ## ...
 ## ... (not yet)
@@ -92,52 +182,65 @@ eps <- 10^-(1:18)
 t.m2 <- t(sapply(eps, re.m2ex3, EXPMlist = expmList))
 ## --> 3 error messages from solve(V), 5 error messages from try(. "R_Eigen" ...)
 
+showProc.time()
+
 cbind(sort(apply(log(t.m2),2, median, na.rm=TRUE)))
 ## 'na.rm=TRUE' needed for Eigen which blows up for the last 3 eps
 t.m2.ranks <- sort(rowMeans(apply(t.m2, 1, rank)))
 cbind(signif(t.m2.ranks, 3))
-## sPs.H08. 2.08
-## sPs.H08b 2.08
-## Ward     2.25
-## s.T.s    5.44
-## s.T.sO   5.44
-## s.P.s    6.06
-## s.P.sO   6.06
-## hybrid   7.25
-## Eigen    8.33
+## lynne (x86_64, Linux 3.14.4-100; Intel i7-4765T), 2014-09:
+## sPs.H08.   2.67
+## sPs.H08b   2.67
+## s.P.sRBS   3.06
+## Ward       4.03
+## AmHi09.13  4.33 <<- still not close to H08 !
+## AmHi09.12  5.86
+## s.T.s      8.33
+## s.T.sO     8.33
+## s.P.s      9.11
+## s.P.sO     9.11
+## hybrid    10.80
+## AmHi09.10 11.70 << astonishingly bad
+## Eigen     12.60
+## AmHi09.08 13.10
+## AmHi09.06 14.40
 
 print(t.m2[, names(t.m2.ranks)[1:8]], digits = 3)
 ## ==> 1st class: H08 (both) and (but slightly better than)  Ward
 ##     2nd class  s.T.s and s.P.s
 ##    "bad" : hybrid and Eigen
 
+## ??? AmHi09 - methods, up to order = 10 are worse !
 if(require(RColorBrewer)) {
     ## Bcol <- brewer.pal(ncol(t.m2),"Dark2")
-    Bcol <- brewer.pal(ncol(t.m2),"Set1")
+    Bcol <- brewer.pal(min(9, ncol(t.m2)), "Set1")
+    Bcol <- Bcol[sqrt(colSums(col2rgb(Bcol)^2)) < 340]
+    ## FIXME: more colors ==> ~/R/MM/GRAPHICS/color-palettes.R
 } else {
     ## 7 from Dark2
     ## Bcol <- c("#1B9E77", "#D95F02", "#7570B3", "#E7298A",
     ##		 "#66A61E", "#E6AB02", "#A6761D")
     ## Rather: those from "Set1"
     Bcol <- c("#E41A1C", "#377EB8", "#4DAF4A",
-	      "#984EA3", "#FF7F00", "#FFFF33",
+	      "#984EA3", "#FF7F00", # too bright: "#FFFF33",
 	      "#A65628", "#F781BF", "#999999")
 }
 
-matplot(eps, t.m2, type = "b", log = "xy", col=Bcol, lty = 1:9, pch=1:9,
+matplot(eps, t.m2, type = "b", log = "xy", col=Bcol, lty = 1:9, pch=1:15,
         axes=FALSE, frame = TRUE,
         xlab = expression(epsilon), ylab = "relative error",
         main = expression(expm(A, method == "*") *"  relative errors for  " *
             A == bgroup("[", atop({-1} *"  "* 1, {epsilon^2} *"  "*{-1}), "]")))
-legend("bottomright",colnames(t.m2),       col=Bcol, lty = 1:7, pch=1:7,
+legend("bottomright",colnames(t.m2),       col=Bcol, lty = 1:9, pch=1:15,
        inset = 0.02)
 if(require("sfsmisc")) {
     sfsmisc::eaxis(1, labels=FALSE)
     sfsmisc::eaxis(1, at = eps[c(TRUE,FALSE)])
-    sfsmisc::eaxis(2, labels=FALSE)
-    op <- par(las=2)
-    sfsmisc::eaxis(2, at = axTicks(2,log=TRUE)[c(TRUE,FALSE,FALSE)])
-    par(op)
+    sfsmisc::eaxis(2)
+    ## sfsmisc::eaxis(2, labels=FALSE)
+    ## op <- par(las=2)
+    ## sfsmisc::eaxis(2, at = axTicks(2,log=TRUE)[c(TRUE,FALSE,FALSE)])
+    ## par(op)
 } else {
     axis(1)
     axis(2)
@@ -151,7 +254,7 @@ me$expA * exp(1) ## the correct value ; numerically identical to simple matrix:
 stopifnot(all.equal(me$expA * exp(1),
 		    rbind(c(  1,  1),
 			  c(ep^2, 1)),
-		    tol = 1e-14))
+		    tolerance = 1e-14))
 ## The relative error (matrices):
 lapply(expmList, function(EXPM) 1 - EXPM(me$A)/me$expA)
 
@@ -166,7 +269,7 @@ round(nDig, 2)
 
 set.seed(17)
 m <- rnilMat(10)
-as(m, "sparseMatrix")# for nicer printing
+(m. <- as(m,"sparseMatrix"))# for nicer printing - and more *below*
 E.m <- expm::expm(m, method="Pade")
 as(E.m, "sparseMatrix")
 
@@ -190,12 +293,13 @@ EmN <- matrix(c(dN, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 Em.xct <- EmN / dN
 
 stopifnot(all.equal(E.m, Em.xct,
-                    check.attributes = FALSE, tol= 1e-13))
-re.x <- sapply(expmL.wo.E, function(EXPM) relErr(Em.xct, EXPM(m)))
+                    check.attributes = FALSE, tolerance= 1e-13))
+re.x <- sapply(expmList, function(EXPM) relErr(Em.xct, EXPM(m)))
 ## with error message from "safe.Eigen"  -->  Eigen is NA here
 
-## result depends quite a bit on platform:
-options(digits = 4, width=90)
+## result depends quite a bit on platform
+which(is.na(re.x))
+(re.x <- re.x[!is.na(re.x)])
 
 ## Pentium-M 32-bit ubuntu gave
 ##      Ward     s.P.s    s.P.sO  sPs.H08.  sPs.H08b     s.T.s    s.T.sO    hybrid
@@ -214,10 +318,19 @@ options(digits = 4, width=90)
 ##     Ward    s.P.s   s.P.sO sPs.H08. sPs.H08b    s.T.s   s.T.sO hybrid
 ## 5.13e-17 3.99e-17 3.99e-17 1.84e-15 1.84e-15 8.44e-17 8.44e-17 5.13e-17
 
-which(is.na(re.x))
-(re.x <- re.x[!is.na(re.x)])
+## 2014-09: AmHi09 are very good (64bit: 8e-17) for p >= 8 (p=6 has 1.5e-11)
 
+not.09.06 <- which(names(re.x) != "AmHi09.06")
 stopifnot(re.x[c("Ward", "s.T.s", "s.T.sO")] < 3e-16,
-          re.x < 1e-13)# <- 32-bit needed 0.451e-14
+          re.x[["AmHi09.06"]] < 9e-11, # x64 & 686(lnx): = 1.509e-11
+          re.x[not.09.06] < 4e-13)# max: 686(32b): 4.52e-14, x64(lnx): 1.103e-16
 
-cat('Time elapsed: ',(p3 <- proc.time())-p2,'\n') # for ``statistical reasons''
+##-- Looking at *sparse* matrices: [C,Fortran "dense" code based methods will fail]:
+(meths <- eval(formals(expm)$method))
+ems <- sapply(meths, function(met)
+              tryCatch(expm::expm(m., method=met), error=identity))
+ok <- !sapply(ems, is, class="error")
+meths[ok] # only two, for now
+## "Higham08" "R_Pade"
+
+showProc.time()
